--Clone Stamp effect filter v1.1 adapted by khaver

obs = obslua
bit = require("bit")

SETTING_SHAPE = 'shape'
SETTING_SIZE = 'size'
SETTING_SOFT = 'softness'
SETTING_INTERP = 'interp'
SETTING_HAIR = 'hair'
SETTING_SPX = 'spx'
SETTING_SPY = 'spy'
SETTING_ASPECT = 'aspect'
SETTING_DPX = 'dpx'
SETTING_DPY = 'dpy'
SETTING_BLEND = 'blend'
SETTING_DRED = 'dred'
SETTING_DGREEN = 'dgreen'
SETTING_DBLUE = 'dblue'
SETTING_GAMMA = 'gamma'

TEXT_SHAPE = 'Shape'
TEXT_SIZE = 'Size'
TEXT_SOFT = 'Softness'
TEXT_INTERP = 'Interpolation'
TEXT_HAIR = 'Show Destination and Source'
TEXT_SPX = 'Source X'
TEXT_SPY = 'Source Y'
TEXT_ASPECT = 'Aspect ratio x:1'
TEXT_DPX = 'Destination X'
TEXT_DPY = 'Destination Y'
TEXT_BLEND = 'Blend Opacity'
TEXT_DRED = 'Red correction'
TEXT_DGREEN = 'Green correction'
TEXT_DBLUE = 'Blue correction'
TEXT_GAMMA = 'Gamma correction'

source_def = {}
source_def.id = 'filter-clonestamp'
source_def.type = obs.OBS_SOURCE_TYPE_FILTER
source_def.output_flags = bit.bor(obs.OBS_SOURCE_VIDEO)

function script_description()
  return [[Clone Stamp v1.1 effect filter adapted by khaver from a Lightworks user effect:
https://github.com/fx-planet/lwks-fx-bundle/blob/master/Repair%20tools/CloneStamp.fx]]
end

function script_load(settings)
  obs.obs_register_source(source_def)
end

function set_render_size(filter)
    target = obs.obs_filter_get_target(filter.context)

    local width, height
    if target == nil then
        width = 0
        height = 0
    else
        width = obs.obs_source_get_base_width(target)
        height = obs.obs_source_get_base_height(target)
    end

    filter.width = width
    filter.height = height
end

source_def.get_name = function()
    return "Clone Stamp"
end

source_def.create = function(settings, source)

    filter = {}
    filter.params = {}
    filter.context = source

    set_render_size(filter)

    obs.obs_enter_graphics()
    filter.effect = obs.gs_effect_create(shader, nil, nil)
    if filter.effect ~= nil then
        filter.params.shape = obs.gs_effect_get_param_by_name(filter.effect, 'shape')
        filter.params.size = obs.gs_effect_get_param_by_name(filter.effect, 'size')
        filter.params.softness = obs.gs_effect_get_param_by_name(filter.effect, 'softness')
        filter.params.interp = obs.gs_effect_get_param_by_name(filter.effect, 'interp')
        filter.params.hair = obs.gs_effect_get_param_by_name(filter.effect, 'hair')
        filter.params.spx = obs.gs_effect_get_param_by_name(filter.effect, 'spx')
        filter.params.spy = obs.gs_effect_get_param_by_name(filter.effect, 'spy')
        filter.params.aspect = obs.gs_effect_get_param_by_name(filter.effect, 'aspect')
        filter.params.dpx = obs.gs_effect_get_param_by_name(filter.effect, 'dpx')
        filter.params.dpy = obs.gs_effect_get_param_by_name(filter.effect, 'dpy')
        filter.params.blend = obs.gs_effect_get_param_by_name(filter.effect, 'blend')
        filter.params.dred = obs.gs_effect_get_param_by_name(filter.effect, 'dred')
        filter.params.dgreen = obs.gs_effect_get_param_by_name(filter.effect, 'dgreen')
        filter.params.dblue = obs.gs_effect_get_param_by_name(filter.effect, 'dblue')
        filter.params.gamma = obs.gs_effect_get_param_by_name(filter.effect, 'gamma')
        filter.params.width = obs.gs_effect_get_param_by_name(filter.effect, 'width')
        filter.params.height = obs.gs_effect_get_param_by_name(filter.effect, 'height')
    end
    obs.obs_leave_graphics()

    if filter.effect == nil then
        source_def.destroy(filter)
        return nil
    end

    source_def.update(filter, settings)
    return filter
end

source_def.destroy = function(filter)
    if filter.effect ~= nil then
        obs.obs_enter_graphics()
        obs.gs_effect_destroy(filter.effect)
        obs.obs_leave_graphics()
    end
	filter = nil
end

source_def.get_width = function(filter)
    return filter.width
end

source_def.get_height = function(filter)
    return filter.height
end

source_def.update = function(filter, settings)
	local theshape = obs.obs_data_get_string(settings, 'sShape')
	local iShape = 0
	if (theshape == 'Ellipe') then iShape = 0 end
	if (theshape == 'Rectangle') then iShape = 1 end
    filter.shape = iShape --obs.obs_data_get_int(settings, SETTING_SHAPE)
    filter.size = obs.obs_data_get_double(settings, SETTING_SIZE)
    filter.softness = obs.obs_data_get_double(settings, SETTING_SOFT)
	local theinterp = obs.obs_data_get_string(settings, 'sInterp')
	local iInterp = 0
	if (theinterp == 'Linear') then iInterp = 0 end
	if (theinterp == 'Square') then iInterp = 1 end
	if (theinterp == 'Sinusoidal') then iInterp = 2 end
    filter.interp = iInterp --obs.obs_data_get_int(settings, SETTING_INTERP)
    filter.hair = obs.obs_data_get_bool(settings, SETTING_HAIR)
    filter.spx = obs.obs_data_get_double(settings, SETTING_SPX)
    filter.spy = obs.obs_data_get_double(settings, SETTING_SPY)
    filter.aspect = obs.obs_data_get_double(settings, SETTING_ASPECT)
    filter.dpx = obs.obs_data_get_double(settings, SETTING_DPX)
    filter.dpy = obs.obs_data_get_double(settings, SETTING_DPY)
    filter.blend = obs.obs_data_get_double(settings, SETTING_BLEND)
    filter.dred = obs.obs_data_get_double(settings, SETTING_DRED)
    filter.dgreen = obs.obs_data_get_double(settings, SETTING_DGREEN)
    filter.dblue = obs.obs_data_get_double(settings, SETTING_DBLUE)
    filter.gamma = obs.obs_data_get_double(settings, SETTING_GAMMA)

    set_render_size(filter)
end

source_def.video_render = function(filter, effect)
  if not obs.obs_source_process_filter_begin(filter.context, obs.GS_RGBA, obs.OBS_NO_DIRECT_RENDERING) then
	obs.obs_source_skip_video_filter(filter.context)
	return
  end

    obs.gs_effect_set_int(filter.params.shape, filter.shape)
    obs.gs_effect_set_float(filter.params.size, filter.size)
    obs.gs_effect_set_float(filter.params.softness, filter.softness)
    obs.gs_effect_set_int(filter.params.interp, filter.interp)
    obs.gs_effect_set_bool(filter.params.hair, filter.hair)
    obs.gs_effect_set_float(filter.params.spx, filter.spx)
    obs.gs_effect_set_float(filter.params.spy, filter.spy)
    obs.gs_effect_set_float(filter.params.aspect, filter.aspect)
    obs.gs_effect_set_float(filter.params.dpx, filter.dpx)
    obs.gs_effect_set_float(filter.params.dpy, filter.dpy)
    obs.gs_effect_set_float(filter.params.blend, filter.blend)
    obs.gs_effect_set_float(filter.params.dred, filter.dred)
    obs.gs_effect_set_float(filter.params.dgreen, filter.dgreen)
    obs.gs_effect_set_float(filter.params.dblue, filter.dblue)
    obs.gs_effect_set_float(filter.params.gamma, filter.gamma)
	obs.gs_effect_set_int(filter.params.width, filter.width)
	obs.gs_effect_set_int(filter.params.height, filter.height)

	obs.gs_blend_state_push()
	obs.gs_blend_function(obs.GS_BLEND_ONE, obs.GS_BLEND_INVSRCALPHA)
    obs.obs_source_process_filter_end(filter.context, filter.effect, filter.width, filter.height)
	obs.gs_blend_state_pop()
end

source_def.get_properties = function(settings)
    props = obs.obs_properties_create()
	local shp = obs.obs_properties_add_list(props, "sShape", TEXT_SHAPE, obs.OBS_COMBO_TYPE_LIST, obs.OBS_COMBO_FORMAT_STRING)
	obs.obs_property_list_add_string(shp, 'Ellipse', 'Ellipse')
	obs.obs_property_list_add_string(shp, 'Rectangle', 'Rectangle')
    obs.obs_properties_add_float_slider(props, SETTING_ASPECT, TEXT_ASPECT, 0.3, 3.3333333, 0.0001)
    obs.obs_properties_add_float_slider(props, SETTING_SIZE, TEXT_SIZE, 0.0, 1.0, 0.0001)
    obs.obs_properties_add_float_slider(props, SETTING_SOFT, TEXT_SOFT, 0.0, 1.0, 0.0001)
	local inter = obs.obs_properties_add_list(props, 'sInterp', TEXT_INTERP, obs.OBS_COMBO_TYPE_LIST, obs.OBS_COMBO_FORMAT_STRING)
	obs.obs_property_list_add_string(inter, 'Linear', 'Linear')
	obs.obs_property_list_add_string(inter, 'Square', 'Square')
	obs.obs_property_list_add_string(inter, 'Sinusoidal', 'Sinusoidal')
	obs.obs_properties_add_bool(props, SETTING_HAIR, TEXT_HAIR)
    obs.obs_properties_add_float_slider(props, SETTING_DPX, TEXT_DPX, 0.0, 1.0, 0.0001)
    obs.obs_properties_add_float_slider(props, SETTING_DPY, TEXT_DPY, 0.0, 1.0, 0.0001)
    obs.obs_properties_add_float_slider(props, SETTING_SPX, TEXT_SPX, 0.0, 1.0, 0.0001)
    obs.obs_properties_add_float_slider(props, SETTING_SPY, TEXT_SPY, 0.0, 1.0, 0.0001)
    obs.obs_properties_add_float_slider(props, SETTING_BLEND, TEXT_BLEND, 0.0, 1.0, 0.0001)
    obs.obs_properties_add_float_slider(props, SETTING_DRED, TEXT_DRED, -1.0, 1.0, 0.0001)
    obs.obs_properties_add_float_slider(props, SETTING_DGREEN, TEXT_DGREEN, -1.0, 1.0, 0.0001)
    obs.obs_properties_add_float_slider(props, SETTING_DBLUE, TEXT_DBLUE, -1.0, 1.0, 0.0001)
    obs.obs_properties_add_float_slider(props, SETTING_GAMMA, TEXT_GAMMA, 0.0, 3.0, 0.0001)

    return props
end

source_def.get_defaults = function(settings)
    obs.obs_data_set_default_string(settings, 'sShape', 'Ellipse')
    obs.obs_data_set_default_double(settings, SETTING_SIZE, 0.33)
    obs.obs_data_set_default_double(settings, SETTING_SOFT, 0.5)
    obs.obs_data_set_default_string(settings, 'sInterp', 'Linear')
    obs.obs_data_set_default_bool(settings, SETTING_HAIR, false)
    obs.obs_data_set_default_double(settings, SETTING_SPX, 0.5)
    obs.obs_data_set_default_double(settings, SETTING_SPY, 0.5)
    obs.obs_data_set_default_double(settings, SETTING_ASPECT, 1.0)
    obs.obs_data_set_default_double(settings, SETTING_DPX, 0.7)
    obs.obs_data_set_default_double(settings, SETTING_DPY, 0.7)
    obs.obs_data_set_default_double(settings, SETTING_BLEND, 1.0)
    obs.obs_data_set_default_double(settings, SETTING_DRED, 0.0)
    obs.obs_data_set_default_double(settings, SETTING_DGREEN, 0.0)
    obs.obs_data_set_default_double(settings, SETTING_DBLUE, 0.0)
    obs.obs_data_set_default_double(settings, SETTING_GAMMA, 1.0)
end

source_def.video_tick = function(filter, seconds)
    set_render_size(filter)
end


shader = [[
//Clone Stamp effect filter adapted by khaver

uniform float4x4 ViewProj;
uniform texture2d image;

uniform int shape;
uniform float size;
uniform float softness;
uniform int interp;
uniform bool hair;
uniform float spx;
uniform float spy;
uniform float aspect;
uniform float dpx;
uniform float dpy;
uniform float blend;
uniform float dred;
uniform float dgreen;
uniform float dblue;
uniform float gamma;
uniform int width;
uniform int height;

#define PI      3.14159265
#define PI_AREA 1.27323954

sampler_state textureSampler {
    Filter    = Linear;
    AddressU  = Clamp;
    AddressV  = Clamp;
};

struct VertDataIn {
    float4 pos : POSITION;
    float2 uv  : TEXCOORD0;
};

struct VertDataOut {
    float4 pos : POSITION;
    float2 uv  : TEXCOORD0;
};

VertDataOut VSDefault(VertDataIn v_in)
{
    VertDataOut vert_out;

    vert_out.pos = mul(float4(v_in.pos.xyz, 1.0), ViewProj);
    vert_out.uv  = v_in.uv;
    return vert_out;
}

float4 ps_ellipse (float2 uv)
{
float _OutputAspectRatio = float(width) / float(height);
	float SrcY = 1.0 - spy;
	float Gamma = gamma * 2.2;

   float4 Src = image.Sample(textureSampler, uv);            // get background texture for edge softness

   //Adjust size for circle
   float CircleSize = size * PI_AREA;

   //Adjust aspect ratio
   float2 DestPos = float2 (dpx, 1.0 - dpy);
   float2 DestAspectAdjustedPos = ((uv - DestPos) / (float2 (aspect, _OutputAspectRatio) * CircleSize)) + DestPos;

   float DestDelta = distance (DestAspectAdjustedPos, DestPos);

   //apply effect only in the effect radius
   if (CircleSize <= DestDelta) return Src;

   //correct Softness radius (cannot be greater than the Effect radius)
   float SoftRadius = CircleSize * (1.0 - softness);

   // distance between the softness radius and the pixel position
   float SoftRing = DestDelta - SoftRadius;

   // initiate Softness to set Transparency (0 - fully solid by default)
   float Soft = 0.0;

   // if the pixel is in the soft area, interpolate softness as per Interpolation parameter
   if (SoftRing >= 0.0) {
      SoftRing /= (CircleSize - SoftRadius);

      Soft = (interp == 0) ? SoftRing
           : (interp == 1) ? 1.0 - pow (1.0 - pow (SoftRing, 2.0), 0.5)
                                  : 0.5 - (cos (SoftRing * PI) / 2.0);
   }

   //Offset Source and Destination
   float2 xy = uv + float2 (spx, dpy) - float2 (dpx, spy);

   // get texture for Destination replacement
   float4 Dest = image.Sample(textureSampler, xy);

   //applies color correction
   Dest.rgb += float3 (dred, dgreen, dblue);
   Dest.rgb = pow(Dest.rgb, float3(2.2/Gamma,2.2/Gamma,2.2/Gamma));
   if (hair) Dest.r = saturate(Dest.r + 0.5);

   // apply softness by merging with the background
   Dest = lerp (Dest, Src, Soft);

   // Apply opacity the same way
   return float4 (lerp (Src.rgb, Dest.rgb, blend), Src.a);
}

float4 ps_rectangle (float2 uv)
{
float _OutputAspectRatio = float(width) / float(height);
	float SrcY = 1.0 - spy;
	float Gamma = gamma * 2.2;

   float4 Src = image.Sample(textureSampler, uv);            // get background texture for edge softness

   //get Destination Position - so it can be modified (parameters are constant, not variables)
   float2 DestPos  = float2 (dpx, 1.0 - dpy);
   float2 DestSize = float2 (aspect, _OutputAspectRatio) * size;
   float2 SoftSize = DestSize * (1.0 - softness);

   //define box effect limits
   float2 BoxMin = DestPos - DestSize / 2.0;
   float2 BoxMax = BoxMin + DestSize;

   //apply effect only in the effect radius
   //if (any ((uv - BoxMin) < float2(0.0,0.0)) || any ((uv - BoxMax) > float2(0.0,0.0))) return Src;
   if ((uv.x - BoxMin.x) < 0.0 || (uv.y - BoxMin.y) < 0.0 || (uv.x - BoxMax.x) > 0.0 || (uv.y - BoxMax.y) > 0.0) return Src;

   //define softness effect limits
   float2 SoftMin = DestPos - SoftSize / 2.00;
   float2 SoftMax = SoftMin + SoftSize;

   //define softness range
   float2 RangeMin = (uv - SoftMin) / (BoxMin - SoftMin);
   float2 RangeMax = (uv - SoftMax) / (BoxMax - SoftMax);

   // if the pixel is in the soft area, interpolate softness as per Interpolation parameter

   if (interp == 1) {
      RangeMin = float2(1.0,1.0) - pow ((float2(1.0,1.0) - pow (RangeMin, float2(2.0,2.0))), float2(0.5,0.5));
      RangeMax = float2(1.0,1.0) - pow ((float2(1.0,1.0) - pow (RangeMax, float2(2.0,2.0))), float2(0.5,0.5));
   }
   else if (interp == 2) {
      RangeMin = float2(0.5,0.5) - (cos (RangeMin * PI) / 2.0);
      RangeMax = float2(0.5,0.5) - (cos (RangeMax * PI) / 2.0);
   }

   RangeMin = float2(1.0,1.0) - RangeMin;
   RangeMax = float2(1.0,1.0) - RangeMax;

   float Soft_1 = ((uv.x >= BoxMin.x) && (uv.x <= SoftMin.x)) ? RangeMin.x : 1.0;
   float Soft_2 = ((uv.y >= BoxMin.y) && (uv.y <= SoftMin.y)) ? RangeMin.y : 1.0;

   if ((uv.x <= BoxMax.x) && (uv.x >= SoftMax.x)) Soft_1 = min (Soft_1, RangeMax.x);
   if ((uv.y <= BoxMax.y) && (uv.y >= SoftMax.y)) Soft_2 = min (Soft_2, RangeMax.y);

   float Soft = saturate (min (Soft_1, Soft_2) * Soft_1 * Soft_2);

   //Offset Source and Destination
   float2 xy = uv + float2 (spx, dpy) - float2 (dpx, spy);

   // get texture for Destination replacement
   float4 Dest = image.Sample(textureSampler, xy);

   //applies color correction
   Dest.rgb += float3 (dred, dgreen, dblue);
   Dest.rgb = pow(Dest.rgb, float3(2.2/Gamma,2.2/Gamma,2.2/Gamma));
   if (hair) Dest.r = saturate(Dest.r + 0.5);

   // apply softness by merging with the background
   Dest = lerp (Src, Dest, Soft);

   // Apply opacity the same way
   return float4 (lerp (Src.rgb, Dest.rgb, blend), Src.a);
}



float4 PassThrough(VertDataOut v_in) : TARGET
{
	float2 pixel = float2(1.0 / float(width), 1.0 / float(height));
	float SrcY = 1.0 - spy;
    float2 xy = v_in.uv;
	if (hair) {
		if ((xy.x >= (spx - pixel.x) && xy.x <= (spx + pixel.x)) || (xy.y >= (SrcY - pixel.y) && xy.y <= (SrcY + pixel.y))) return float4(0.0, 1.0, 0.0, 1.0);
	}


   if (shape == 0) return ps_ellipse(xy);
   return ps_rectangle(xy);

}

technique Draw
{
    pass
    {
        vertex_shader = VSDefault(v_in);
        pixel_shader  = PassThrough(v_in);
    }
}
]]
